#!/usr/bin/env python3
"""
cli.py

Main CLI entry point, now using:
  - Command mapping for dispatch
  - Short and long flags for --file (-f), --dir (-d), --recursive (-R), etc.
  - Uniform file/directory handling
  - Optional recursive discovery of pyproject.toml files
"""

import argparse
import os
import sys
from typing import List

# Local modules
from soliloquy import poetry_ops
from soliloquy import version_ops
from soliloquy import remote_ops
from soliloquy import eval_ops
from soliloquy import pyproject_ops
from soliloquy import test_ops

def resolve_targets(directory: str = None,
                    file: str = None,
                    recursive: bool = False,
                    default_filename: str = "pyproject.toml") -> List[str]:
    """
    Determine which pyproject.toml file(s) to operate on.

    - If `file` is specified, return just that file (if it exists).
    - Else if `directory` is specified:
        - If `recursive=True`, walk the directory and include every subdir
          containing a file named `default_filename`.
        - Otherwise, just return one file: <directory>/<default_filename>
    """
    if file:
        file_path = os.path.abspath(file)
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"Specified file does not exist: {file_path}")
        return [file_path]

    if directory:
        directory_path = os.path.abspath(directory)
        if not os.path.isdir(directory_path):
            raise NotADirectoryError(f"Specified directory does not exist: {directory_path}")

        if recursive:
            matched = []
            for root, dirs, files in os.walk(directory_path):
                if default_filename in files:
                    matched.append(os.path.join(root, default_filename))
            if not matched:
                raise FileNotFoundError(
                    f"No '{default_filename}' files found recursively in {directory_path}"
                )
            return matched
        else:
            single_file = os.path.join(directory_path, default_filename)
            if not os.path.isfile(single_file):
                raise FileNotFoundError(f"No {default_filename} found in {directory_path}")
            return [single_file]

    # If neither file nor directory is provided, fallback or raise
    raise ValueError("Must provide either -f/--file or -d/--dir")


#------------------------------------------------------------------------------
# Command Handlers
#------------------------------------------------------------------------------

def handle_lock(args):
    try:
        targets = resolve_targets(args.dir, args.file, args.recursive)
        for t in targets:
            location = os.path.dirname(t)
            poetry_ops.poetry_lock(location)
    except Exception as e:
        print(f"Error in lock command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_install(args):
    try:
        targets = resolve_targets(args.dir, args.file, args.recursive)
        for t in targets:
            location = os.path.dirname(t)
            poetry_ops.poetry_install(
                location,
                extras=args.extras,
                with_dev=args.dev,
                all_extras=args.all_extras
            )
    except Exception as e:
        print(f"Error in install command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_build(args):
    # We see which sub-subcommand was chosen
    if args.build_cmd == "pkg":
        handle_build_pkg(args)
    elif args.build_cmd == "mono":
        handle_build_mono(args)
    else:
        print("Unknown build subcommand:", args.build_cmd)
        sys.exit(1)

def handle_build_pkg(args):
    """
    Handles 'cli build pkg' logic.
    Supports -f, -d, and optional -R.
    """
    from soliloquy import poetry_ops
    try:
        # This calls your existing "recursive_build" or a variation of it.
        poetry_ops.recursive_build(
            file=args.file,
            directory=args.dir,
            recursive=args.recursive
        )
    except Exception as e:
        print(f"Error in build pkg command: {e}", file=sys.stderr)
        sys.exit(1)

def handle_build_mono(args):
    """
    Handles 'cli build mono' logic.
    Only supports -f (a single aggregator pyproject).
    Now also supports optional --build-dir and --no-cleanup.
    """
    from soliloquy import poetry_ops
    try:
        # Pass our new arguments:
        poetry_ops.build_monorepo(
            file=args.file,
            build_dir=args.build_dir,
            cleanup=not args.no_cleanup
        )
    except Exception as e:
        print(f"Error in build mono command: {e}", file=sys.stderr)
        sys.exit(1)



def handle_version(args):
    # We renamed the argument to --file/-f for consistency
    try:
        version_ops.bump_or_set_version(
            pyproject_file=args.file,
            bump=args.bump,
            set_ver=args.set_ver
        )
    except Exception as e:
        print(f"Error in version command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_remote(args):
    if args.remote_cmd == "fetch":
        try:
            ver = remote_ops.fetch_remote_pyproject_version(
                git_url=args.git_url,
                branch=args.branch,
                subdirectory=args.subdir
            )
            if ver:
                print(f"Fetched remote version: {ver}")
            else:
                print("Failed to fetch remote version.")
        except Exception as e:
            print(f"Error in remote fetch command: {e}", file=sys.stderr)
            sys.exit(1)

    elif args.remote_cmd == "update":
        try:
            success = remote_ops.remote_update_bulk(
            file=args.file,
            directory=args.dir,
            recursive=args.recursive,
            output=args.output
        )
            if not success:
                sys.exit(1)
        except Exception as e:
            print(f"Error in remote update command: {e}", file=sys.stderr)
            sys.exit(1)



def handle_test(args):
    """
    Enhanced 'test' command handler that supports:
      --mode=monorepo | each | single
    """
    try:
        if args.mode == "monorepo":
            # Run a single pytest invocation for the entire monorepo
            test_ops.run_monorepo_tests(
                root_dir=args.dir,
                num_workers=args.num_workers
            )

        elif args.mode == "each":
            # Discover subpackages and run each individually
            test_ops.run_subpackage_tests(
                root_dir=args.dir,
                num_workers=args.num_workers
            )

        else:  # "single" (the default)
            # The existing logic: run tests in `args.dir` or current dir
            poetry_ops.run_pytests(
                test_directory=args.dir,
                num_workers=args.num_workers
            )

    except Exception as e:
        print(f"Error in test command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_analyze(args):
    try:
        eval_ops.analyze_test_file(
            file_path=args.file,
            required_passed=args.required_passed,
            required_skipped=args.required_skipped
        )
    except Exception as e:
        print(f"Error in analyze command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_pyproject(args):
    try:
        # For "pyproject" subcommand, we interpret --pyproject as a directory or file input
        # If you prefer short flags here, you can add -f or -d similarly
        targets = resolve_targets(directory=args.pyproject, file=None, recursive=False)
        pyproject_path = targets[0]

        print("Extracting dependencies from pyproject.toml ...")
        paths = pyproject_ops.extract_path_dependencies(pyproject_path)
        if paths:
            print("Local (path) dependencies:")
            print(", ".join(paths))
        else:
            print("No local path dependencies found.")

        git_deps = pyproject_ops.extract_git_dependencies(pyproject_path)
        if git_deps:
            print("\nGit dependencies:")
            for name, details in git_deps.items():
                print(f"{name}: {details}")
        else:
            print("No Git dependencies found.")

        if args.update_version:
            print(f"\nUpdating local dependency versions to {args.update_version} ...")
            pyproject_ops.update_dependency_versions(pyproject_path, args.update_version)
    except Exception as e:
        print(f"Error in pyproject command: {e}", file=sys.stderr)
        sys.exit(1)


def handle_publish(args):
    try:
        poetry_ops.poetry_publish(
            file=args.file,
            directory=args.dir,
            recursive=args.recursive,
            username=args.username,
            password=args.password
        )
    except Exception as e:
        print(f"Error in publish command: {e}", file=sys.stderr)
        sys.exit(1)

def handle_lint(args):
    """
    Dispatch linting tasks to poetry_ops.poetry_ruff_lint.
    """
    try:
        # If the user provided -d/--dir, we'll pass that to poetry_ruff_lint
        directory = args.dir or "."
        poetry_ops.poetry_ruff_lint(directory=directory, fix=args.fix)
    except Exception as e:
        print(f"Error in lint command: {e}", file=sys.stderr)
        sys.exit(1)

def handle_version(args):
    """
    Updated version handler that allows -f/--file, -d/--dir, and -R/--recursive.
    Also supports --bump or --set for the version operation.
    """
    from soliloquy.version_ops import version_bulk

    try:
        # We call version_bulk with the right arguments
        version_bulk(
            file=args.file,
            directory=args.dir,
            recursive=args.recursive,
            bump=args.bump,
            set_ver=args.set_ver
        )
    except Exception as e:
        print(f"Error in version command: {e}", file=sys.stderr)
        sys.exit(1)


#------------------------------------------------------------------------------
# Main Entry Point with Subparser & Command Mapping
#------------------------------------------------------------------------------

def main():
    parser = argparse.ArgumentParser(
        description="A CLI for managing a Python monorepo with multiple standalone packages."
    )
    subparsers = parser.add_subparsers(dest="command", required=True, help="Available commands")

    # lock
    lock_parser = subparsers.add_parser("lock", help="Generate a poetry.lock file")
    lock_parser.add_argument("-d", "--dir", type=str, help="Directory containing pyproject.toml")
    lock_parser.add_argument("-f", "--file", type=str, help="Explicit path to a pyproject.toml file")
    lock_parser.add_argument("-R", "--recursive", action="store_true", help="Recursively find pyproject.toml files")
    
    # install
    install_parser = subparsers.add_parser("install", help="Install dependencies")
    install_parser.add_argument("-d", "--dir", type=str, help="Directory containing pyproject.toml")
    install_parser.add_argument("-f", "--file", type=str, help="Explicit path to a pyproject.toml file")
    install_parser.add_argument("-R", "--recursive", action="store_true", help="Recursively find pyproject.toml files")
    install_parser.add_argument("--extras", type=str, help="Extras to include (e.g. 'full')")
    install_parser.add_argument("--dev", action="store_true", help="Include dev dependencies")
    install_parser.add_argument("--all-extras", action="store_true", help="Include all extras")
    
    #-----------------------------------------------------------------
    # build
    #-----------------------------------------------------------------
    build_parser = subparsers.add_parser("build", help="Build packages")
    build_subparsers = build_parser.add_subparsers(dest="build_cmd", required=True)

    # build pkg
    pkg_parser = build_subparsers.add_parser("pkg", help="Build a single or multiple packages")
    pkg_parser.add_argument("-f", "--file", type=str, help="Path to a pyproject.toml")
    pkg_parser.add_argument("-d", "--dir", type=str, help="Directory containing pyproject.toml(s)")
    pkg_parser.add_argument("-R", "--recursive", action="store_true",
                            help="Recursively find pyproject.toml files in directory mode")
    # The function that handles 'build pkg'
    pkg_parser.set_defaults(func=handle_build_pkg)

    # build mono
    mono_parser = build_subparsers.add_parser("mono", help="Build monorepo aggregator only (requires -f).")
    mono_parser.add_argument("-f", "--file", type=str, required=True,
                             help="Path to a monorepo aggregator pyproject.toml (with package-mode=false)")
    mono_parser.add_argument("--build-dir", type=str,
                             help="Directory to clone and build Git dependencies (default: temp dir).")
    mono_parser.add_argument("--no-cleanup", action="store_true",
                             help="If set, do NOT remove the build directory (only applies if we create a temp dir).")
    mono_parser.set_defaults(func=handle_build_mono)

    
    # version
    version_parser = subparsers.add_parser("version", help="Bump or set package version(s)")
    version_parser.add_argument("-f", "--file", help="Path to a specific pyproject.toml")
    version_parser.add_argument("-d", "--dir", help="Directory containing one or more pyproject.toml")
    version_parser.add_argument("-R", "--recursive", action="store_true", help="Recursively find pyproject.toml files")

    vgroup = version_parser.add_mutually_exclusive_group(required=True)
    vgroup.add_argument("--bump", choices=["major", "minor", "patch", "finalize"],
                        help="Bump the version (e.g. patch, major, minor, finalize)")
    vgroup.add_argument("--set", dest="set_ver", help="Explicit version to set (e.g. 2.0.0.dev1)")

    version_parser.set_defaults(func=handle_version)

    
    # remote
    remote_parser = subparsers.add_parser("remote", help="Remote operations for Git dependencies")
    remote_subparsers = remote_parser.add_subparsers(dest="remote_cmd", required=True)
    
    fetch_parser = remote_subparsers.add_parser("fetch", help="Fetch version from remote GitHub pyproject.toml")
    fetch_parser.add_argument("--git-url", type=str, required=True, help="GitHub repository URL")
    fetch_parser.add_argument("--branch", type=str, default="main", help="Branch name (default: main)")
    fetch_parser.add_argument("--subdir", type=str, default="", help="Subdirectory for pyproject.toml")
    
    update_parser = remote_subparsers.add_parser("update", help="Update local pyproject.toml with remote versions")
    update_parser.add_argument("-f", "--file", help="Path to the local pyproject.toml")
    update_parser.add_argument("-d", "--dir", help="Directory containing pyproject.toml(s)")
    update_parser.add_argument("-R", "--recursive", action="store_true", help="Recursively find pyproject.toml files")
    update_parser.add_argument("-o", "--output", help="Optional output file path (applies only if a single file is found)")


    # test
    test_parser = subparsers.add_parser("test", help="Run tests using pytest")
    test_parser.add_argument("-d", "--dir", type=str, default=".", help="Directory to run tests (default: .)")
    test_parser.add_argument("--num-workers", type=int, default=1, help="Number of parallel workers")
    test_parser.add_argument(
        "--mode",
        choices=["single", "monorepo", "each"],
        default="single",
        help="Testing mode: single package, entire monorepo, or each subpackage. (default: single)"
    )
    
    # analyze
    analyze_parser = subparsers.add_parser("analyze", help="Analyze test results from a JSON file")
    analyze_parser.add_argument("-f", "--file", help="Path to the JSON file with test results", required=True)
    analyze_parser.add_argument("--required-passed", type=str, help="Threshold for passed tests (e.g. 'gt:75')")
    analyze_parser.add_argument("--required-skipped", type=str, help="Threshold for skipped tests (e.g. 'lt:20')")

    # pyproject
    pyproject_parser = subparsers.add_parser("pyproject", help="Operate on pyproject.toml dependencies")
    pyproject_parser.add_argument("--pyproject", required=True, help="Path (or directory) to pyproject.toml")
    pyproject_parser.add_argument("--update-version", type=str, help="Update local dependency versions")

    # publish
    publish_parser = subparsers.add_parser("publish", help="Publish package(s) to PyPI")
    publish_parser.add_argument("-f", "--file", type=str, help="Path to a specific pyproject.toml")
    publish_parser.add_argument("-d", "--dir", type=str, help="Directory containing pyproject.toml(s)")
    publish_parser.add_argument("-R", "--recursive", action="store_true", help="Recursively find pyproject.toml files")
    publish_parser.add_argument("--username", type=str, help="PyPI username")
    publish_parser.add_argument("--password", type=str, help="PyPI password")


    # --------------------------------------------------
    # lint
    # --------------------------------------------------
    lint_parser = subparsers.add_parser("lint", help="Run Ruff lint checks (and optionally autofix).")
    lint_parser.add_argument("-d", "--dir", type=str,
                             help="Directory to lint (default: current directory).")
    lint_parser.add_argument("--fix", action="store_true",
                             help="Apply autofixes using 'ruff check --fix'.")

    # Map subcommands to handler functions
    COMMANDS = {
        "lock": handle_lock,
        "install": handle_install,
        "build": handle_build,
        "version": handle_version,
        "remote": handle_remote,
        "test": handle_test,
        "analyze": handle_analyze,
        "pyproject": handle_pyproject,
        "publish": handle_publish,
        "lint": handle_lint
    }

    args = parser.parse_args()
    handler = COMMANDS.get(args.command)
    if handler is None:
        parser.print_help()
        sys.exit(1)

    handler(args)


if __name__ == "__main__":
    main()



# soliloquy/ops/poetry_utils.py

import subprocess
import sys
from typing import List, Optional

def run_command(cmd: List[str], cwd: Optional[str] = None) -> int:
    """
    Runs the given command (list of strings) in the specified working directory.
    Returns the integer exit code (0 indicates success).

    Args:
        cmd: A list of command and arguments, e.g. ["poetry", "build"].
        cwd: An optional directory in which to run the command. Defaults to None (current dir).

    Example:
        exit_code = run_command(["poetry", "install"], cwd="/path/to/package")
        if exit_code != 0:
            print("Command failed!")
    """
    # Print the command for debugging
    print(f"[poetry_utils] Running command: {' '.join(cmd)} (cwd={cwd or '.'})", flush=True)
    try:
        process = subprocess.run(
            cmd,
            cwd=cwd,
            text=True,
            capture_output=True,
            shell=False
        )
        return process.returncode
    except Exception as e:
        print(f"[poetry_utils] Error running command {cmd}: {e}", file=sys.stderr)
        return 1


#!/usr/bin/env python3
"""
version_ops.py

Provides functions to:
  - Read the current version from pyproject.toml.
  - Bump the current version (major, minor, patch) or finalize a dev release.
  - Validate that a user-provided new version is not lower than the current one.
  - Update the pyproject.toml with the new version.

Intended for use in a unified monorepo management CLI.
"""

import sys
from packaging.version import Version, InvalidVersion
from tomlkit import parse, dumps
from .pyproject_ops import find_pyproject_files

def read_pyproject_version(file_path):
    """
    Reads the current version from the provided pyproject.toml file.

    Args:
        file_path (str): Path to the pyproject.toml file.

    Returns:
        tuple: A tuple containing the current version string and the
               tomlkit Document representing the file.
    Raises:
        KeyError: If the version key is missing.
    """
    try:
        with open(file_path, "r") as f:
            content = f.read()
    except Exception as e:
        print(f"Error reading {file_path}: {e}", file=sys.stderr)
        sys.exit(1)
    
    doc = parse(content)
    try:
        version = doc["tool"]["poetry"]["version"]
    except KeyError:
        raise KeyError("No version found under [tool.poetry] in the given pyproject.toml")
    return version, doc

def bump_version(current_version, bump_type):
    """
    Bumps the current version up using semantic versioning.
    Supports:
      - Bumping stable versions (major, minor, patch) which also start a dev cycle.
      - Bumping within a dev cycle.
      - Finalizing a dev version (removing the .dev suffix).

    Args:
        current_version (str): The current version (e.g. "1.0.0" or "1.0.1.dev2").
        bump_type (str): One of "major", "minor", "patch", or "finalize".

    Returns:
        str: The new version string.
    Raises:
        ValueError: If the current version is invalid or the bump operation cannot be performed.
    """
    try:
        ver = Version(current_version)
    except InvalidVersion as e:
        raise ValueError(f"Invalid current version '{current_version}': {e}")

    # Check if it's a dev release
    is_dev = ver.dev is not None
    major, minor, patch = ver.release

    if bump_type == "finalize":
        if is_dev:
            # Remove the dev segment
            new_version = f"{major}.{minor}.{patch}"
        else:
            raise ValueError("Current version is stable; nothing to finalize.")
    elif bump_type == "major":
        major += 1
        minor = 0
        patch = 0
        new_version = f"{major}.{minor}.{patch}.dev1"
    elif bump_type == "minor":
        minor += 1
        patch = 0
        new_version = f"{major}.{minor}.{patch}.dev1"
    elif bump_type == "patch":
        if is_dev:
            # Increment the dev counter if already in a dev cycle.
            new_dev = ver.dev + 1
            new_version = f"{major}.{minor}.{patch}.dev{new_dev}"
        else:
            patch += 1
            new_version = f"{major}.{minor}.{patch}.dev1"
    else:
        raise ValueError("bump_type must be one of: 'major', 'minor', 'patch', or 'finalize'")

    return new_version


def validate_and_set_version(current_version, new_version):
    """
    Validates that the new version is not lower than the current version.

    Args:
        current_version (str): The current version string.
        new_version (str): The target version string.

    Returns:
        str: The new version if it is valid.
    Raises:
        ValueError: If new_version is lower than current_version.
    """
    try:
        cur_ver = Version(current_version)
        tgt_ver = Version(new_version)
    except InvalidVersion as e:
        raise ValueError(f"Invalid version provided: {e}")

    if tgt_ver < cur_ver:
        raise ValueError("You cannot bump the version downwards. The target version must be higher than the current version.")
    
    return new_version


def update_pyproject_version(file_path, new_version):
    """
    Updates the pyproject.toml file with the new version.

    Args:
        file_path (str): The path to the pyproject.toml file.
        new_version (str): The new version string.
    
    Returns:
        None
    """
    try:
        current_version, doc = read_pyproject_version(file_path)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)

    # Update the version field if it exists
    if "tool" in doc and "poetry" in doc["tool"]:
        doc["tool"]["poetry"]["version"] = new_version
    else:
        print(f"Error: Invalid pyproject.toml structure in {file_path}.", file=sys.stderr)
        sys.exit(1)

    try:
        with open(file_path, "w") as f:
            f.write(dumps(doc))
    except Exception as e:
        print(f"Error writing updated pyproject.toml: {e}", file=sys.stderr)
        sys.exit(1)

    print(f"Bumped version from {current_version} to {new_version} in {file_path}.")


def bump_or_set_version(pyproject_file, bump=None, set_ver=None):
    """
    Executes either a version bump or a direct version set on the given pyproject.toml file.
    
    Args:
        pyproject_file (str): Path to the pyproject.toml file.
        bump (str, optional): The type of bump ("major", "minor", "patch", or "finalize").
        set_ver (str, optional): A specific version string to set.
    
    Returns:
        None
    """
    try:
        current_version, _ = read_pyproject_version(pyproject_file)
    except Exception as e:
        print(f"Error reading current version: {e}", file=sys.stderr)
        sys.exit(1)

    try:
        if bump:
            new_version = bump_version(current_version, bump)
        elif set_ver:
            new_version = validate_and_set_version(current_version, set_ver)
        else:
            print("No version operation specified.", file=sys.stderr)
            sys.exit(1)
    except ValueError as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)

    update_pyproject_version(pyproject_file, new_version)

def version_bulk(
    file: str = None,
    directory: str = None,
    recursive: bool = False,
    bump: str = None,
    set_ver: str = None
):
    """
    Finds one or more pyproject.toml files (using -f / -d / -R logic),
    then applies either a bump or set-version operation to each.

    :param file: Explicit path to a pyproject.toml
    :param directory: Path to a directory that has (or contains) pyproject.toml
    :param recursive: If True, search subdirectories for pyproject.toml
    :param bump: One of "major", "minor", "patch", "finalize"
    :param set_ver: Explicit version string to set
    """
    try:
        pyproject_files = find_pyproject_files(file, directory, recursive)
    except (FileNotFoundError, NotADirectoryError, ValueError) as e:
        print(f"Error discovering pyproject files: {e}", file=sys.stderr)
        sys.exit(1)

    if not bump and not set_ver:
        print("No version operation specified (bump or set_ver).", file=sys.stderr)
        sys.exit(1)

    # For each discovered pyproject, apply bump or set
    for pyproj_file in pyproject_files:
        print(f"\n---\nProcessing {pyproj_file} ...")
        try:
            bump_or_set_version(pyproj_file, bump=bump, set_ver=set_ver)
        except SystemExit as e:
            # If bump_or_set_version sys.exit's on an error, handle or re-raise
            print(f"Aborting due to error on {pyproj_file}.", file=sys.stderr)
            sys.exit(e.code)
        except Exception as e:
            print(f"Error modifying version in {pyproj_file}: {e}", file=sys.stderr)
            sys.exit(1)

    print("\nAll specified pyproject.toml files have been processed.")


import os
import sys
import tomlkit
from .poetry_ops import run_command
from .pyproject_ops import extract_path_dependencies, extract_git_dependencies, find_pyproject_files

def recursive_build(file: str = None, directory: str = None, recursive: bool = False):
    """
    Recursively build packages based on path and/or git dependencies
    extracted from one or more pyproject.toml files.

    Steps:
      1. Identify which pyproject.toml files to process.
      2. Parse each pyproject.toml, checking 'package-mode'.
      3. If package-mode=true (or unspecified), build the root package itself.
      4. Gather path dependencies & build each local path package.
      5. Gather git dependencies (optional) and decide whether to build them.
    """
    # Step 1: Find all relevant pyproject.toml files
    try:
        pyproject_files = find_pyproject_files(
            file=file, directory=directory, recursive=recursive
        )
    except Exception as e:
        print(f"Error discovering pyproject files: {e}", file=sys.stderr)
        sys.exit(1)

    # Step 2: For each discovered pyproject.toml
    for pyproject_path in pyproject_files:
        print(f"\nExamining '{pyproject_path}' ...")

        # Parse the doc using tomlkit
        try:
            with open(pyproject_path, "r", encoding="utf-8") as f:
                doc = tomlkit.parse(f.read())
        except Exception as e:
            print(f"Error reading {pyproject_path}: {e}", file=sys.stderr)
            continue  # or sys.exit(1)

        # Retrieve the [tool.poetry] table
        tool_poetry = doc.get("tool", {}).get("poetry", {})
        project_dir = os.path.dirname(pyproject_path)

        # Determine if it's package-mode or aggregator-mode
        # Default to 'True' if not specified (so it's buildable).
        is_package_mode = tool_poetry.get("package-mode", True)
        if isinstance(is_package_mode, bool):
            pass  # Good, we can use it directly
        else:
            # If someone used a string "false", we can interpret that
            if str(is_package_mode).lower() == "false":
                is_package_mode = False
            else:
                is_package_mode = True  # fallback

        package_name = tool_poetry.get("name")
        package_version = tool_poetry.get("version")

        if is_package_mode:
            # We consider it a normal package, so build the root package first
            print(f"Detected normal package '{package_name}' (version: {package_version}).")
            try:
                run_command(["poetry", "build"], cwd=project_dir)
            except Exception as e:
                print(f"Failed to build root package in {project_dir}: {e}", file=sys.stderr)
        else:
            # It's aggregator (monorepo) mode
            print("Detected 'monorepo aggregator' (package-mode=false). Skipping direct build of root.")

        # Step 4: Build local path dependencies
        path_deps = extract_path_dependencies(pyproject_path)
        if path_deps:
            print(f"Found {len(path_deps)} path dependencies in {pyproject_path}.")
            for sub_path in path_deps:
                full_path = os.path.join(project_dir, sub_path)
                sub_pyproject = os.path.join(full_path, "pyproject.toml")

                if os.path.isdir(full_path) and os.path.isfile(sub_pyproject):
                    print(f"Building local path dependency: {full_path}")
                    try:
                        run_command(["poetry", "build"], cwd=full_path)
                    except Exception as e:
                        print(f"Failed to build path dependency {full_path}: {e}", file=sys.stderr)
                else:
                    print(f"Skipping {full_path}: not a valid local package directory.")
        else:
            print(f"No local path dependencies found in {pyproject_path}.")

        # Step 5 (Optional): Gather and build Git dependencies
        git_deps = extract_git_dependencies(pyproject_path)
        if git_deps:
            print(f"Found {len(git_deps)} Git dependencies in {pyproject_path}.")

            # If you want to build them locally, you'd clone them and run "poetry build".
            # Otherwise, skip. Example pseudocode:
            for name, details in git_deps.items():
                git_url = details.get("git")
                if not git_url:
                    continue
                print(f"Detected git dependency '{name}' from {git_url}. (Skipping or cloning...)")
                # ...
        else:
            print(f"No Git dependencies found in {pyproject_path}.")



# soliloquy/ops/remote_ops.py

import os
import sys
import requests
from typing import Optional, List

from urllib.parse import urljoin
from tomlkit import parse, dumps, inline_table

# If you have a function for discovering relevant pyproject.toml files:
from soliloquy.ops.pyproject_ops import find_pyproject_files

def fetch_remote_pyproject_version(
    git_url: str,
    branch: str = "main",
    subdirectory: str = ""
) -> Optional[str]:
    """
    Attempts to fetch a remote pyproject.toml (on GitHub) and parse out
    the [tool.poetry].version field.

    - Only handles GitHub repos at the moment.
    - If subdirectory is provided, it is appended to the raw URL path.

    Returns:
        The version string (e.g. '1.2.3') if found, else None.
    """
    if "github.com" not in git_url:
        print(f"[remote_ops] Currently only supports GitHub: {git_url}", file=sys.stderr)
        return None

    # Remove trailing .git if present
    repo_path = git_url.split("github.com/")[-1]
    if repo_path.endswith(".git"):
        repo_path = repo_path[:-4]

    # Construct the raw URL to the pyproject.toml
    # e.g. https://raw.githubusercontent.com/<user>/<repo>/<branch>/<subdirectory>/pyproject.toml
    base_url = f"https://raw.githubusercontent.com/{repo_path}/{branch}/"
    if subdirectory and not subdirectory.endswith("/"):
        subdirectory += "/"
    pyproject_url = urljoin(base_url, f"{subdirectory}pyproject.toml")

    print(f"[remote_ops] Fetching remote pyproject from {pyproject_url}")
    try:
        resp = requests.get(pyproject_url)
        resp.raise_for_status()
    except Exception as e:
        print(f"[remote_ops] Failed to fetch {pyproject_url}: {e}", file=sys.stderr)
        return None

    try:
        doc = parse(resp.text)
        version = doc.get("tool", {}).get("poetry", {}).get("version")
        if not version:
            print(f"[remote_ops] No version found in remote pyproject.toml at {pyproject_url}", file=sys.stderr)
            return None
        return version
    except Exception as e:
        print(f"[remote_ops] Could not parse remote TOML: {e}", file=sys.stderr)
        return None


def update_pyproject_with_versions(file_path: str) -> Optional[str]:
    """
    Reads a local pyproject.toml, and for each Git-based dependency:
      - fetches the remote version (if possible),
      - updates that dependency's version to ^<remote_version>,
      - preserves original 'optional' status or extras membership.

    Returns:
      The updated TOML string (if changes made or not),
      or None if an error occurred (like invalid structure).
    """
    if not os.path.isfile(file_path):
        print(f"[remote_ops] File not found: {file_path}", file=sys.stderr)
        return None

    try:
        with open(file_path, "r", encoding="utf-8") as f:
            doc = parse(f.read())
    except Exception as e:
        print(f"[remote_ops] Error reading {file_path}: {e}", file=sys.stderr)
        return None

    # Locate [tool.poetry] section
    tool_poetry = doc.get("tool", {}).get("poetry")
    if not tool_poetry:
        print(f"[remote_ops] Invalid pyproject structure in {file_path}. Missing [tool.poetry].", file=sys.stderr)
        return None

    dependencies = tool_poetry.get("dependencies", {})
    # If [tool.poetry.extras] section exists, store it for references
    extras_section = tool_poetry.get("extras", {})
    had_extras = bool(extras_section)

    def dep_in_extras(dep_name: str) -> bool:
        # Checks if dep_name is listed in any extras array
        for extra_deps in extras_section.values():
            if dep_name in extra_deps:
                return True
        return False

    # For each dependency, if it has "git", we attempt to fetch version
    updated_any = False

    for dep_name, details in list(dependencies.items()):
        if not (isinstance(details, dict) and "git" in details):
            # not a git-based dependency
            continue

        git_url = details["git"]
        branch = details.get("branch", "main")
        subdir = details.get("subdirectory", "")
        print(f"\n[remote_ops] Checking Git dep '{dep_name}' -> {git_url}@{branch} subdir='{subdir}'")

        # Keep track of if it was originally optional
        originally_optional = bool(details.get("optional", False)) or dep_in_extras(dep_name)

        remote_ver = fetch_remote_pyproject_version(git_url, branch, subdir)
        if not remote_ver:
            # If we fail to fetch a remote version, we won't force a change. 
            # You could optionally mark them optional if you want, or skip entirely.
            print(f"  [remote_ops] Could not get remote version for '{dep_name}'. Skipping update.")
            continue

        print(f"  [remote_ops] Fetched remote version: {remote_ver}")

        # Build a new inline table with ^version
        new_inline = inline_table()
        new_inline["version"] = f"^{remote_ver}"

        # Preserve optional if it was originally optional
        if originally_optional:
            new_inline["optional"] = True

        # We also preserve other keys like 'git', 'branch', 'subdirectory' if you want:
        new_inline["git"] = git_url
        if "branch" in details:
            new_inline["branch"] = details["branch"]
        if "subdirectory" in details:
            new_inline["subdirectory"] = details["subdirectory"]

        # Update the dictionary
        dependencies[dep_name] = new_inline
        updated_any = True

    # If no changes, just return the original content as TOML string
    if not updated_any:
        print(f"[remote_ops] No updates applied in {file_path}")
        return dumps(doc)

    # Optionally clean extras so that each extra only references existing dependencies
    if had_extras:
        for extra_name, extra_list in extras_section.items():
            # remove anything that doesn't exist in dependencies now
            extras_section[extra_name] = [dep for dep in extra_list if dep in dependencies]
        tool_poetry["extras"] = extras_section  # reassign

    tool_poetry["dependencies"] = dependencies
    return dumps(doc)


def update_and_write_pyproject(input_file_path: str, output_file_path: Optional[str] = None) -> bool:
    """
    Reads a pyproject.toml from input_file_path, updates Git deps with remote versions,
    writes the result back to 'output_file_path' or overwrites the original.

    Returns True on success (even if no changes), False on error.
    """
    updated_doc_str = update_pyproject_with_versions(input_file_path)
    if updated_doc_str is None:
        return False

    # If no output path, overwrite the original
    target = output_file_path or input_file_path
    try:
        with open(target, "w", encoding="utf-8") as f:
            f.write(updated_doc_str)
        print(f"[remote_ops] Updated file written to {target}")
        return True
    except Exception as e:
        print(f"[remote_ops] Error writing updated pyproject to {target}: {e}", file=sys.stderr)
        return False


def remote_update_bulk(
    file: Optional[str] = None,
    directory: Optional[str] = None,
    recursive: bool = False,
    output: Optional[str] = None
) -> bool:
    """
    Bulk-update Git-based dependencies in one or more pyproject.toml files.
    
    - If 'file' is provided, updates that one file.
    - Else if 'directory' is provided:
        * If 'recursive=True', find all pyproject.toml in subdirectories.
        * Otherwise, just that single 'directory/pyproject.toml' (if present).
    - 'output': If exactly one pyproject file is found, you can direct the updated
      content to a new path. If multiple are found but 'output' is set, 
      we log a warning and just update each in-place.

    Returns:
      True if all updates are successful, False if any file fails to update.
    """
    try:
        pyprojects = find_pyproject_files(file=file, directory=directory, recursive=recursive)
    except Exception as e:
        print(f"[remote_ops] Error discovering pyproject files: {e}", file=sys.stderr)
        return False

    if not pyprojects:
        print("[remote_ops] No pyproject files found to update.")
        return True  # or False, depending on what you consider "success"

    if output and len(pyprojects) > 1:
        print("[remote_ops] Multiple pyproject files found but a single --output was provided. "
              "We will update each file in-place instead of writing to a single output.",
              file=sys.stderr)
        out_path = None
    else:
        out_path = output

    overall_success = True

    for idx, pyproj_path in enumerate(pyprojects, start=1):
        print(f"\n[remote_ops] ({idx}/{len(pyprojects)}) Updating remote deps in {pyproj_path}")
        success = update_and_write_pyproject(
            input_file_path=pyproj_path,
            output_file_path=out_path if len(pyprojects) == 1 else None
        )
        if not success:
            overall_success = False

    if overall_success:
        print("[remote_ops] Remote update completed successfully for all files.")
    else:
        print("[remote_ops] Some updates failed.", file=sys.stderr)

    return overall_success



#!/usr/bin/env python3
"""
eval_ops.py

Provides functions to analyze test result data from a JSON file.

The JSON file is expected to have:
  - A "summary" section with keys: "total", "passed", "failed", "skipped".
  - A "tests" list, where each test contains an "outcome" (e.g., "passed", "failed", "skipped")
    and a "keywords" list for tags.
    
The module:
  - Reads the JSON file.
  - Prints a summary table with counts and percentages.
  - Groups tests by tags (excluding unwanted tags such as "tests", tags starting with "test_", tags
    ending with "_test.py", or empty tags).
  - Checks threshold conditions for passed and skipped percentages (if provided) and exits with an error
    code if the conditions are not satisfied.
"""

import json
import sys
from collections import defaultdict
import argparse


def evaluate_threshold(value, threshold):
    """
    Evaluate if the given value meets the specified threshold condition.

    The threshold format should be: operator:limit (e.g., "gt:50").
    Supported operators:
      - gt: greater than
      - lt: less than
      - eq: equal to
      - ge: greater than or equal to
      - le: less than or equal to

    Returns:
        bool: True if the condition is met, False otherwise.
    """
    try:
        op, limit = threshold.split(":")
        limit = float(limit)
    except ValueError as e:
        raise ValueError(
            f"Invalid threshold format '{threshold}'. Expected format: 'gt:<number>' etc."
        ) from e

    if op == "gt":
        return value > limit
    elif op == "lt":
        return value < limit
    elif op == "eq":
        return value == limit
    elif op == "ge":
        return value >= limit
    elif op == "le":
        return value <= limit
    else:
        raise ValueError(
            f"Invalid operator '{op}'. Use one of: 'gt', 'lt', 'eq', 'ge', 'le'."
        )


def analyze_test_file(file_path, required_passed=None, required_skipped=None):
    """
    Analyzes a JSON file with test results.

    The function:
      - Prints a summary table with the total count and percentage for each outcome.
      - Checks if the percentage of passed or skipped tests meet the specified thresholds.
      - Groups tests by tags (excluding tags that are deemed irrelevant).
      - Prints detailed tag-based results.

    If thresholds are not met, the function exits with an error.

    Args:
        file_path (str): Path to the JSON file.
        required_passed (str, optional): Threshold for passed tests (e.g., "gt:50").
        required_skipped (str, optional): Threshold for skipped tests (e.g., "lt:20").
    """
    try:
        with open(file_path, "r") as f:
            data = json.load(f)
    except FileNotFoundError:
        print(f"Error: File not found: {file_path}")
        sys.exit(1)
    except json.JSONDecodeError:
        print(f"Error: Could not decode JSON from {file_path}")
        sys.exit(1)
    except Exception as e:
        print(f"Unexpected error reading {file_path}: {e}")
        sys.exit(1)

    summary = data.get("summary", {})
    tests = data.get("tests", [])
    if not summary or not tests:
        print("No test data or summary found in the provided file.")
        sys.exit(1)

    total_tests = summary.get("total", 0)
    print("\nTest Results Summary:")
    print(f"{'Category':<15}{'Count':<10}{'Total':<10}{'% of Total':<10}")
    print("-" * 50)
    for category in ["passed", "skipped", "failed"]:
        count = summary.get(category, 0)
        percentage = (count / total_tests) * 100 if total_tests > 0 else 0
        print(f"{category.capitalize():<15}{count:<10}{total_tests:<10}{percentage:<10.2f}")

    # Calculate percentages for threshold evaluation.
    passed_pct = (summary.get("passed", 0) / total_tests) * 100 if total_tests > 0 else 0
    skipped_pct = (summary.get("skipped", 0) / total_tests) * 100 if total_tests > 0 else 0

    threshold_error = False
    if required_passed and not evaluate_threshold(passed_pct, required_passed):
        print(
            f"\nWARNING: Passed percentage ({passed_pct:.2f}%) does not meet the condition '{required_passed}'!"
        )
        threshold_error = True

    if required_skipped and not evaluate_threshold(skipped_pct, required_skipped):
        print(
            f"WARNING: Skipped percentage ({skipped_pct:.2f}%) does not meet the condition '{required_skipped}'!"
        )
        threshold_error = True

    # Group tests by tags.
    tag_outcomes = {}
    for test in tests:
        outcome = test.get("outcome", "").lower()
        for tag in test.get("keywords", []):
            # Exclude unwanted tags.
            if tag == "tests" or tag.startswith("test_") or tag.endswith("_test.py") or tag.strip() == "":
                continue
            if tag not in tag_outcomes:
                tag_outcomes[tag] = {"passed": 0, "skipped": 0, "failed": 0, "total": 0}
            tag_outcomes[tag]["total"] += 1
            if outcome == "passed":
                tag_outcomes[tag]["passed"] += 1
            elif outcome == "skipped":
                tag_outcomes[tag]["skipped"] += 1
            elif outcome == "failed":
                tag_outcomes[tag]["failed"] += 1

    print("\nTag-Based Results:")
    header = f"{'Tag':<30}{'Passed':<10}{'Skipped':<10}{'Failed':<10}{'Total':<10}{'% Passed':<10}{'% Skipped':<10}{'% Failed':<10}"
    print(header)
    print("-" * len(header))
    # Sort tags by percentage passed descending then alphabetically.
    sorted_tags = sorted(
        tag_outcomes.items(),
        key=lambda item: (
            -(item[1]["passed"] / item[1]["total"] * 100 if item[1]["total"] > 0 else 0),
            item[0],
        ),
    )
    for tag, outcomes in sorted_tags:
        total = outcomes["total"]
        passed_pct = (outcomes["passed"] / total * 100) if total > 0 else 0
        skipped_pct = (outcomes["skipped"] / total * 100) if total > 0 else 0
        failed_pct = (outcomes["failed"] / total * 100) if total > 0 else 0
        print(
            f"{tag:<30}{outcomes['passed']:<10}{outcomes['skipped']:<10}{outcomes['failed']:<10}"
            f"{total:<10}{passed_pct:<10.2f}{skipped_pct:<10.2f}{failed_pct:<10.2f}"
        )

    # If thresholds are not met, exit with a non-zero status code.
    if threshold_error:
        sys.exit(1)
    else:
        print("\nTest analysis completed successfully.")





# soliloquy/phases/prepare.py

import sys
from typing import Any

# Import the operations from their respective modules
from soliloquy.ops.version_ops import bulk_bump_or_set_version
from soliloquy.ops.lint_ops import run_ruff_lint
from soliloquy.ops.git_ops import git_commit_all_changes

def run_prepare(args: Any) -> None:
    """
    The 'prepare' phase:

      1. Bump or set version (recursively if requested)
      2. Lint + autofix
      3. Git commit changes

    This function expects to be called with an 'args' object that has:
      - args.file (optional, for a single pyproject.toml)
      - args.directory (optional, for a directory of packages or aggregator)
      - args.recursive (bool)
      - args.bump (string or None)
      - args.set_ver (string or None)
      - args.commit_msg (string)
      - possibly other lint-related flags if needed
    """

    # --------------------------------------------------
    # 1. Bump or set version
    # --------------------------------------------------
    if not args.bump and not args.set_ver:
        print("[prepare] No version operation specified. Skipping version update.")
    else:
        bulk_bump_or_set_version(
            file=args.file,
            directory=args.directory,
            recursive=args.recursive,
            bump=args.bump,
            set_ver=args.set_ver
        )

    # --------------------------------------------------
    # 2. Lint + autofix
    # --------------------------------------------------
    # For demonstration, let's lint exactly one directory. 
    # If you want to lint multiple subdirectories, adapt accordingly.
    lint_dir = args.directory or "."
    # If you want to exit on lint errors, set exit_on_error=True below.
    # Here, let's keep it false so we can proceed to commit anyway.
    run_ruff_lint(directories=[lint_dir], fix=True, exit_on_error=False)

    # --------------------------------------------------
    # 3. Git commit
    # --------------------------------------------------
    commit_msg = getattr(args, "commit_msg", "chore: prepare changes")
    success = git_commit_all_changes(commit_msg)
    if not success:
        print("[prepare] Git commit failed or there were no changes to commit.")
        # Decide whether to sys.exit(1) or just log and proceed
        # sys.exit(1)

    print("[prepare] Complete.")



# soliloquy/ops/test_ops.py

import os
import sys
import subprocess
import tempfile
import shutil
from typing import Dict, List, Union, Tuple

import tomlkit

from soliloquy.ops.pyproject_ops import (
    find_pyproject_files,
    extract_path_dependencies,
    extract_git_dependencies
)
from soliloquy.ops.poetry_utils import run_command  # For uniform command calls, if desired


def run_pytests(
    test_directory: str = ".",
    num_workers: int = 1
) -> Dict[str, Union[bool, int, str]]:
    """
    Runs pytest (via 'poetry run pytest') in the given directory.
    If num_workers > 1, runs pytest with '-n <num_workers>' for parallel tests.

    Returns a dict with:
      {
        "success": bool,        # True if returncode == 0
        "returncode": int,      # The actual exit code from pytest
        "directory": str,       # The directory tested
      }
    """
    cmd = ["poetry", "run", "pytest"]
    if num_workers > 1:
        cmd.extend(["-n", str(num_workers)])
    print(f"[test_ops] Running tests in {test_directory} -> {' '.join(cmd)}")

    # Here we directly call subprocess, but you could also use run_command(...) if desired.
    proc = subprocess.run(cmd, cwd=test_directory)
    rc = proc.returncode

    return {
        "success": (rc == 0),
        "returncode": rc,
        "directory": test_directory,
    }


def run_tests_with_mode(
    file: str = None,
    directory: str = None,
    recursive: bool = False,
    mode: str = "single",
    num_workers: int = 1,
    cleanup: bool = True
) -> Dict[str, Union[bool, List[Dict[str, Union[bool, int, str]]]]]:
    """
    Entry point to run tests according to a 'mode'.

    Args:
        file: Path to a single pyproject.toml (standalone).
        directory: Directory with packages or aggregator.
        recursive: If True, find multiple pyprojects recursively (used for fallback).
        mode: "single", "monorepo", or "each".
        num_workers: Parallel workers for pytest (-n).

    Returns:
      {
        "success": bool,            # True if all tests pass
        "details": List[dict],      # A list of test result dicts from run_pytests
      }
    """
    if mode not in ("single", "monorepo", "each"):
        print(f"[test_ops] Invalid test mode: {mode}", file=sys.stderr)
        return {"success": False, "details": []}

    # ------------------------------------------------------
    # Discover pyprojects
    # ------------------------------------------------------
    try:
        pyprojects = find_pyproject_files(file, directory, recursive)
    except Exception as e:
        print(f"[test_ops] Error finding pyproject files: {e}", file=sys.stderr)
        return {"success": False, "details": []}

    # If no pyproject found, fallback to single test in the directory
    if not pyprojects:
        print("[test_ops] No pyproject.toml found. Running single test in the specified directory.")
        test_dir = directory or "."
        result = run_pytests(test_dir, num_workers)
        return {"success": result["success"], "details": [result]}

    # We'll consider the first pyproject as "primary" unless mode=each.
    primary_pyproj = pyprojects[0]
    aggregator, pkg_name = _check_if_aggregator(primary_pyproj)

    # ------------------------------------------------------
    # Mode: single
    # ------------------------------------------------------
    if mode == "single":
        # Just run 1 test in the directory (like a standalone approach)
        test_dir = directory or os.path.dirname(primary_pyproj)
        result = run_pytests(test_dir, num_workers)
        return {"success": result["success"], "details": [result]}

    # ------------------------------------------------------
    # Mode: monorepo
    # ------------------------------------------------------
    elif mode == "monorepo":
        # Run a single test from the top-level directory (like one big combined run)
        test_dir = directory or os.path.dirname(primary_pyproj)
        result = run_pytests(test_dir, num_workers)
        return {"success": result["success"], "details": [result]}

    # ------------------------------------------------------
    # Mode: each
    # ------------------------------------------------------
    else:
        # aggregator => gather path deps & git deps, test each
        # else => fallback to scanning multiple pyprojects
        all_passed = True
        details: List[Dict[str, Union[bool, int, str]]] = []

        if aggregator:
            print(f"[test_ops] Aggregator detected ('package-mode=false'), testing each path dependency individually.")

            # 1) Test local path dependencies
            deps = extract_path_dependencies(primary_pyproj)
            base_dir = os.path.dirname(primary_pyproj)

            for dep_rel_path in deps:
                subpkg_path = os.path.join(base_dir, dep_rel_path)
                if not os.path.isdir(subpkg_path):
                    print(f"  Skipping invalid subpackage dir: {subpkg_path}", file=sys.stderr)
                    all_passed = False
                    continue
                r = run_pytests(subpkg_path, num_workers)
                details.append(r)
                if not r["success"]:
                    all_passed = False

            # NEW: 2) Test Git-based dependencies
            git_deps = extract_git_dependencies(primary_pyproj)
            if git_deps:
                # We'll clone them into a single temp folder,
                # run tests in each subdirectory that has a pyproject.toml.
                git_ok = _test_git_deps(git_deps, details, num_workers, cleanup)
                if not git_ok:
                    all_passed = False

        else:
            # Non-aggregator => we test each discovered pyproject in subdirs individually
            print("[test_ops] No aggregator found. We'll test each discovered pyproject in subdirs individually.")
            for pyproj in pyprojects:
                proj_dir = os.path.dirname(pyproj)
                r = run_pytests(proj_dir, num_workers)
                details.append(r)
                if not r["success"]:
                    all_passed = False

        return {"success": all_passed, "details": details}


def _check_if_aggregator(pyproj_path: str) -> (bool, str):
    """
    Parses 'pyproj_path' with tomlkit and checks if [tool.poetry.package-mode] is False.
    Returns:
      (is_aggregator, package_name)
    """
    with open(pyproj_path, "r", encoding="utf-8") as f:
        doc = tomlkit.parse(f.read())
    poetry_table = doc.get("tool", {}).get("poetry", {})
    pkg_name = poetry_table.get("name", "unknown")

    pkg_mode = poetry_table.get("package-mode", True)
    if isinstance(pkg_mode, str):
        # if "false" as string => aggregator
        pkg_mode = (pkg_mode.lower() != "false")

    is_aggregator = (pkg_mode is False)
    return is_aggregator, pkg_name


# --------------------------------------------------------------------------
# NEW: Test Git-based dependencies (similar to building them in build_monorepo)
# --------------------------------------------------------------------------
def _test_git_deps(
    git_deps: dict,
    details: List[Dict[str, Union[bool, int, str]]],
    num_workers: int,
    cleanup: bool = True,  # NEW
) -> bool:
    """
    Clones each Git-based dependency, then runs tests (by:
      1) 'poetry install' in that subdirectory (optional),
      2) 'poetry run pytest' in that subdirectory.

    If cleanup=False, we do NOT remove the temporary directory,
    enabling the caller to reuse it for subsequent steps.

    Returns True if all tests pass, otherwise False.
    """
    import tempfile
    import shutil
    from collections import defaultdict

    # group by (git_url, branch)
    grouped = defaultdict(list)
    for dep_name, config in git_deps.items():
        git_url = config.get("git")
        branch = config.get("branch", "main")
        subdir = config.get("subdirectory", ".")
        grouped[(git_url, branch)].append((dep_name, subdir))

    if not grouped:
        return True

    print(f"[test_ops] Found {len(git_deps)} Git dependencies to test. Cloning & testing each subdir ...")

    all_passed = True
    build_root = tempfile.mkdtemp(prefix="mono_test_")
    print(f"[test_ops] Using temporary directory for Git-based tests: {build_root}")

    try:
        for (git_url, branch), sub_pkgs in grouped.items():
            print(f"\n[test_ops] Cloning {git_url} (branch: {branch}) to test {len(sub_pkgs)} subdir(s).")
            clone_dir_name = branch.replace("/", "-") + "_clone"
            clone_dir = os.path.join(build_root, clone_dir_name)

            clone_cmd = [
                "git", "clone",
                "--depth", "1",
                "--branch", branch,
                git_url,
                clone_dir
            ]
            rc = run_command(clone_cmd)
            if rc != 0:
                print(f"  [test_ops] Failed to clone {git_url}. Skipping these sub-pkgs.", file=sys.stderr)
                all_passed = False
                continue

            # For each subdirectory in that repo, do a test run
            for (dep_name, subdir) in sub_pkgs:
                test_path = os.path.join(clone_dir, *subdir.split("/"))  # ensure cross-platform
                pyproj_file = os.path.join(test_path, "pyproject.toml")
                if not os.path.isdir(test_path):
                    print(f"  [test_ops][{dep_name}] Subdirectory '{subdir}' not found.", file=sys.stderr)
                    all_passed = False
                    continue
                if not os.path.isfile(pyproj_file):
                    print(f"  [test_ops][{dep_name}] No pyproject.toml in {test_path}. Skipping tests.", file=sys.stderr)
                    all_passed = False
                    continue

                # Now run tests
                print(f"  [test_ops][{dep_name}] Running tests in {test_path}")
                test_result = run_pytests(test_path, num_workers)
                details.append(test_result)
                if not test_result["success"]:
                    all_passed = False

    finally:
        # Only remove the temp directory if cleanup is True
        if cleanup:
            shutil.rmtree(build_root, ignore_errors=True)
            print(f"[test_ops] Cleaned up temporary test directory: {build_root}")
        else:
            print(f"[test_ops] Skipping cleanup of {build_root} (cleanup=False).")

    return all_passed


# soliloquy/phases/release.py

import sys
import os
from typing import Any

from soliloquy.phases.validate import run_validate
from soliloquy.ops.poetry_utils import run_command
from soliloquy.ops.pyproject_ops import find_pyproject_files
from soliloquy.ops.remote_ops import remote_update_bulk
from soliloquy.ops.publish_ops import publish_packages
from soliloquy.ops.test_ops import _check_if_aggregator  # or re-implement aggregator check
from soliloquy.ops.build_ops import build_packages


def run_release(args: Any) -> None:
    """
    The 'release' phase:
      - Validate (test-only, but in "each" mode we get subpackage results)
      - For aggregator + each-mode, build/update/publish only subpackages that passed
      - For monorepo or single-mode, either all pass => build/publish, or fail => exit
      - Remote update (optionally aggregator or subpackages)
    """

    print("[release] Running validation (test only).")
    test_results = run_validate(args)
    test_mode = getattr(args, "test_mode", "single")

    if test_mode != "each":
        # monorepo or single => if success, build + remote update + publish all; else fail
        if not test_results["success"]:
            print("[release] Some tests failed in monorepo/single mode. Exiting.", file=sys.stderr)
            sys.exit(1)

        print("[release] Tests passed. Building all packages ...")
        build_ok = build_packages(
            file=args.file,
            directory=args.directory,
            recursive=args.recursive
        )
        if not build_ok:
            print("[release] Build failed. Exiting.", file=sys.stderr)
            sys.exit(1)

        print("[release] Remote update ...")
        if not remote_update_bulk(
            file=args.file,
            directory=args.directory,
            recursive=args.recursive
        ):
            print("[release] Remote update failed. Exiting.", file=sys.stderr)
            sys.exit(1)

        print("[release] Publishing packages ...")
        repository = getattr(args, "repository", None)
        pub_ok = publish_packages(
            file=args.file,
            directory=args.directory,
            recursive=args.recursive,
            username=getattr(args, "publish_username", None),
            password=getattr(args, "publish_password", None),
            repository=repository
        )
        if not pub_ok:
            print("[release] Publish failed. Exiting.", file=sys.stderr)
            sys.exit(1)

        print("[release] Done.")
        return

    # ------------------------------------------------------------------
    # If test_mode == "each", partial pass scenario:
    # we check aggregator or multiple pyprojects, build/publish only those that pass
    # ------------------------------------------------------------------
    if not test_results["details"]:
        print("[release] No subpackages tested? Exiting.", file=sys.stderr)
        sys.exit(1)

    # We assume the first pyproject is the aggregator if aggregator is True
    pyprojects = find_pyproject_files(args.file, args.directory, args.recursive)
    if not pyprojects:
        print("[release] No pyprojects found, can't proceed.", file=sys.stderr)
        sys.exit(1)

    aggregator_pyproj = pyprojects[0]
    is_agg, agg_name = _check_if_aggregator(aggregator_pyproj)

    # Collect a list of subpackage directories that passed
    passed_dirs = [
        d["directory"] for d in test_results["details"] if d["success"]
    ]
    if not passed_dirs:
        print("[release] All subpackages failed. Nothing to publish.", file=sys.stderr)
        sys.exit(1)

    # Possibly do aggregator-level remote update once if aggregator
    if is_agg:
        print(f"[release] Aggregator '{agg_name}' detected. Doing aggregator-level remote update once ...")
        ru_ok = remote_update_bulk(
            file=aggregator_pyproj,
            directory=None,  # or aggregator's directory
            recursive=False
        )
        if not ru_ok:
            print("[release] Remote update failed. Exiting.", file=sys.stderr)
            sys.exit(1)
    else:
        # If not aggregator, optional. If you have multiple packages each with its own pyproject,
        # you might do remote update on each. We'll do a single recursion for all.
        print("[release] Non-aggregator. Doing a single remote update for all found pyprojects ...")
        ru_ok = remote_update_bulk(
            file=None,
            directory=args.directory,
            recursive=args.recursive
        )
        if not ru_ok:
            print("[release] Remote update failed. Exiting.", file=sys.stderr)
            sys.exit(1)

    # Now for each subpackage that passed, do a build & publish
    # We'll skip aggregator itself, if aggregator is not meant to be published
    overall_publish_success = True
    for sub_dir in passed_dirs:
        # If aggregator is the same dir, skip
        if is_agg and (os.path.abspath(sub_dir) == os.path.dirname(aggregator_pyproj)):
            print(f"[release] Skipping aggregator dir {sub_dir}.")
            continue

        print(f"\n[release] Building subpackage: {sub_dir}")
        build_rc = run_command(["poetry", "build"], cwd=sub_dir)
        if build_rc != 0:
            print(f"[release] Build failed in {sub_dir}", file=sys.stderr)
            overall_publish_success = False
            continue

        print(f"[release] Publishing subpackage: {sub_dir}")
        publish_cmd = ["poetry", "publish"]
        repo = getattr(args, "repository", None)
        if repo:
            publish_cmd.extend(["--repository", repo])
        uname = getattr(args, "publish_username", None)
        pwd = getattr(args, "publish_password", None)
        if uname:
            publish_cmd.extend(["--username", uname])
        if pwd:
            publish_cmd.extend(["--password", pwd])

        pub_rc = run_command(publish_cmd, cwd=sub_dir)
        if pub_rc != 0:
            print(f"[release] Publish failed in {sub_dir}", file=sys.stderr)
            overall_publish_success = False

    if not overall_publish_success:
        print("[release] Some subpackages failed to build or publish.", file=sys.stderr)
        sys.exit(1)

    print("\n[release] Done. All passing subpackages have been published.")



#!/usr/bin/env python3
"""
remote_ops.py

Fixes:
  - Preserves original 'optional' status for Git deps instead of forcing optional=True.
  - Does not forcibly create [tool.poetry.extras] if none existed.
  - If a dep was already optional or was in extras, remains optional. Otherwise, stays non-optional.
  - If version cannot be fetched, do not automatically mark it optional unless it was optional before.

"""

import os
from urllib.parse import urljoin

import requests
from tomlkit import parse, dumps, inline_table

# If you have this function in pyproject_ops, import it:
# from .pyproject_ops import find_pyproject_files
# For demonstration, we'll assume it's imported.


def fetch_remote_pyproject_version(git_url, branch="main", subdirectory=""):
    """
    (unchanged)
    """
    try:
        if "github.com" not in git_url:
            raise ValueError("Only GitHub repositories are supported by this function.")
        
        # Remove trailing .git if present.
        repo_path = git_url.split("github.com/")[1]
        if repo_path.endswith(".git"):
            repo_path = repo_path[:-4]
        
        # Build the raw URL; ensure subdirectory ends with "/" if provided.
        base_url = f"https://raw.githubusercontent.com/{repo_path}/{branch}/"
        if subdirectory and not subdirectory.endswith("/"):
            subdirectory += "/"
        pyproject_url = urljoin(base_url, f"{subdirectory}pyproject.toml")
        
        response = requests.get(pyproject_url)
        response.raise_for_status()
        doc = parse(response.text)
        version = doc.get("tool", {}).get("poetry", {}).get("version")
        if version is None:
            print(f"Version key not found in remote pyproject.toml from {pyproject_url}")
        return version
    except Exception as e:
        print(f"Error fetching pyproject.toml from {git_url}: {e}")
        return None


def update_pyproject_with_versions(file_path):
    """
    Reads the local pyproject.toml file and updates Git-based dependencies by:
      - Fetching the remote version (if possible).
      - Preserving the original 'optional' status unless it was previously optional or listed in extras.
      - Only adding an extras section if one already existed or if we need to preserve an existing one.
    """
    # 1) Load the pyproject.toml
    try:
        with open(file_path, "r") as f:
            content = f.read()
        doc = parse(content)
    except Exception as e:
        print(f"Error reading {file_path}: {e}")
        return None

    # Attempt to retrieve relevant sections
    try:
        tool_section = doc["tool"]
        poetry_section = tool_section["poetry"]
    except KeyError:
        print(f"Error: Invalid pyproject.toml structure in {file_path}.", flush=True)
        return None

    dependencies = poetry_section.get("dependencies", {})
    # We only track 'extras' if it actually exists in the doc:
    #   we'll store a boolean if it was originally present.
    had_extras_section = "extras" in poetry_section
    extras = poetry_section.get("extras", {})

    # Helper function: was the dep in any extras list?
    def dep_is_in_extras(dep_name):
        for _, extra_deps in extras.items():
            if dep_name in extra_deps:
                return True
        return False

    for dep_name, details in dependencies.items():
        # Skip non-dict or dict without 'git' key
        if not (isinstance(details, dict) and "git" in details):
            continue

        # We have a Git-based dependency
        git_url = details["git"]
        branch = details.get("branch", "main")
        subdirectory = details.get("subdirectory", "")
        print(f"\nUpdating Git dependency '{dep_name}':")
        print(f"  Repository: {git_url}")
        print(f"  Branch: {branch}")
        print(f"  Subdirectory: {subdirectory}")

        # Keep track of whether this dependency was optional
        #  either because it was explicitly optional or it was in an extras array
        originally_optional = bool(details.get("optional", False)) or dep_is_in_extras(dep_name)

        remote_version = fetch_remote_pyproject_version(git_url, branch=branch, subdirectory=subdirectory)
        if remote_version:
            print(f"  Fetched version: {remote_version}")
            # Create an inline table
            dep_inline = inline_table()
            dep_inline["version"] = f"^{remote_version}"
            if originally_optional:
                # Only mark optional if it was already optional or in extras
                dep_inline["optional"] = True

            # Preserve any other keys that you want (like 'extras' or 'branch' if you prefer).
            # For example, if you'd like to preserve 'branch':
            # if "branch" in details: dep_inline["branch"] = details["branch"]

            dependencies[dep_name] = dep_inline
        else:
            print(f"  Could not fetch remote version for '{dep_name}'.")
            # If it was originally optional, keep it that way; otherwise do not force optional
            if originally_optional:
                details["optional"] = True
            else:
                details.pop("optional", None)  # Remove optional if set

            dependencies[dep_name] = details

    # Next, we only rewrite the [extras] section if it existed or changed in some relevant way
    # However, you might want to "clean" it so that it only references existing dependencies
    # if it existed in the original doc:
    if had_extras_section:
        # Clean existing extras so that each extra only references valid dependencies
        for extra_name, extra_deps in extras.items():
            extras[extra_name] = [dep for dep in extra_deps if dep in dependencies]
        poetry_section["extras"] = extras  # Reassign
    else:
        # If 'extras' didn't exist originally, do not forcibly create it
        if "extras" in poetry_section:
            del poetry_section["extras"]  # remove if we accidentally created an empty table

    # Update the doc with any changes to dependencies
    poetry_section["dependencies"] = dependencies

    return doc


def update_and_write_pyproject(input_file_path, output_file_path=None):
    """
    Updates the specified pyproject.toml file with resolved versions for Git-based dependencies 
    and writes the updated document to a file.

    Returns True on success, False otherwise.
    """
    updated_doc = update_pyproject_with_versions(input_file_path)
    if updated_doc is None:
        print("Failed to update the pyproject.toml document.")
        return False

    # If no output_file_path is given, overwrite the original
    output_file_path = output_file_path or input_file_path

    try:
        with open(output_file_path, "w") as f:
            f.write(dumps(updated_doc))
        print(f"Updated pyproject.toml written to {output_file_path}")
        return True
    except Exception as e:
        print(f"Error writing updated pyproject.toml: {e}")
        return False


# Example "bulk" function to handle -f/-d/-R logic:
from .pyproject_ops import find_pyproject_files

def remote_update_bulk(file=None, directory=None, recursive=False, output=None) -> None:
    """
    Updates one or more pyproject.toml files by resolving Git-based dependencies
    and fetching remote versions.

    :param file: Explicit path to a pyproject.toml
    :param directory: Path to a directory that has (or contains) pyproject.toml
    :param recursive: If True, search subdirectories for pyproject.toml
    :param output: Optional output file path (applies only if a single file is found)

    We'll:
      - Find the relevant pyproject files
      - For each file, call update_and_write_pyproject()
      - Not forcibly add [tool.poetry.extras] if none existed
      - Only set `optional=true` if it was originally optional or in extras
    """
    try:
        pyproject_files = find_pyproject_files(
            file=file, directory=directory, recursive=recursive
        )
    except (FileNotFoundError, NotADirectoryError, ValueError) as e:
        print(f"Error discovering pyproject files: {e}")
        return

    # If multiple files are found, but user provided a single output path:
    if output and len(pyproject_files) > 1:
        print("Warning: Multiple pyproject files found but only one --output was provided.\n"
              "We'll update each file in-place. If you intended separate outputs, run individually.\n")

    for i, pyproj_file in enumerate(pyproject_files, start=1):
        # Only use 'output' if there's exactly one file
        out_file = output if (output and len(pyproject_files) == 1) else None

        print(f"\n---\n[{i}/{len(pyproject_files)}] Updating remote deps for: {pyproj_file}")
        success = update_and_write_pyproject(pyproj_file, out_file)
        if not success:
            print(f"Failed to update remote deps in {pyproj_file}.")
            # Decide if you want to break or keep going



#!/usr/bin/env python3
"""
poetry_ops.py

Key changes:
  - Removed install_poetry()
  - run_command now uses shell=False and an array of arguments
  - If reading/writing pyproject, switch to tomlkit (if needed)
"""
from typing import List, Optional, Dict, Any, Tuple
import os
import subprocess
import sys
import tempfile
import shutil
import tomlkit
from tomlkit import parse
from .pyproject_ops import extract_path_dependencies, extract_git_dependencies, find_pyproject_files

def run_command(command_args, cwd=None):
    """
    Run a shell command (as a list of args) with shell=False,
    capturing output. Raises CalledProcessError on failure.
    """
    try:
        result = subprocess.run(
            command_args,
            cwd=cwd,
            text=True,
            capture_output=True,
            shell=False
        )
        if result.stdout:
            print(result.stdout, end="")
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        print(f"Error running command: {e.stderr}", file=sys.stderr)
        sys.exit(e.returncode)

def poetry_lock(location):
    """
    Run 'poetry lock' in the specified location (directory).
    """
    print(f"Generating poetry.lock in {location}...")
    run_command(["poetry", "lock"], cwd=location)


def poetry_install(location, extras=None, with_dev=False, all_extras=False):
    """
    Run 'poetry install' in the specified location.
    """
    print(f"Installing dependencies in {location}...")
    command = ["poetry", "install", "--no-cache", "-vv"]
    if all_extras:
        command.append("--all-extras")
    if extras:
        for e in extras.split(","):
            command.extend(["--extras", e.strip()])
    if with_dev:
        command.extend(["--with", "dev"])
    run_command(command, cwd=location)


def extract_path_dependencies(pyproject_path):
    """
    Example reading with tomlkit instead of toml.
    """
    print(f"Extracting path dependencies from {pyproject_path}...")
    try:
        with open(pyproject_path, "r", encoding="utf-8") as f:
            doc = tomlkit.parse(f.read())
    except Exception as e:
        print(f"Error reading {pyproject_path}: {e}", file=sys.stderr)
        sys.exit(1)

    deps = doc.get("tool", {}).get("poetry", {}).get("dependencies", {})
    path_deps = [
        v["path"]
        for v in deps.values()
        if isinstance(v, dict) and "path" in v
    ]
    return path_deps

def run_pytests(test_directory=".", num_workers=1):
    """
    Run pytest in the specified directory.
    """
    command = ["poetry", "run", "pytest"]
    if num_workers > 1:
        command.extend(["-n", str(num_workers)])
    print(f"Running tests in '{test_directory}' with command: {' '.join(command)}")
    run_command(command, cwd=test_directory)


def poetry_ruff_lint(directory=".", fix=False):
    """
    Runs Ruff lint checks in the specified directory.
    If fix=True, also pass --fix to autofix issues.

    :param directory: The directory to lint (defaults to '.')
    :param fix: Whether to run with --fix (autofixing) enabled.
    :param cwd: The current working directory to run the command in (defaults to None)
    """
    # Construct the base command: 'poetry run ruff check <dir>'
    command = ["poetry", "run", "ruff", "check", directory]
    if fix:
        command.append("--fix")

    print(f"Running Ruff on '{directory}' (fix={fix}) with cwd='{directory}'...")
    run_command(command, cwd=directory)


def build_monorepo(
    file: str,
    build_dir: str = None,
    cleanup: bool = True
):
    """
    Build a 'monorepo aggregator' pyproject.toml that has 'package-mode=false'
    by:
      1) Building all local path dependencies
      2) Cloning and building Git-based dependencies
      3) Moving the resulting dist artifacts into build_dir (if provided).

    :param file: Path to the aggregator pyproject.toml
    :param build_dir: If provided, we place cloned Git repos AND final build
                      artifacts here. If None, we use a temp folder for cloning
                      but still place final artifacts there.
    :param cleanup: Whether to remove our build_dir if it’s a temp folder.
    """
    import os
    import tempfile
    import shutil
    from typing import Dict, Tuple, List
    import glob

    from .pyproject_ops import extract_path_dependencies, extract_git_dependencies
    import tomlkit

    aggregator_path = os.path.abspath(file)
    if not os.path.isfile(aggregator_path):
        raise FileNotFoundError(f"Monorepo aggregator file not found: {aggregator_path}")

    print(f"Building monorepo aggregator from: {aggregator_path}")

    with open(aggregator_path, "r", encoding="utf-8") as f:
        doc = tomlkit.parse(f.read())

    tool_poetry = doc.get("tool", {}).get("poetry", {})
    package_mode = tool_poetry.get("package-mode", True)
    if package_mode is not False:
        print("Warning: 'package-mode=false' not found or set. Proceeding as aggregator, but be aware.")

    aggregator_dir = os.path.dirname(aggregator_path)

    # ------------------------------------------------------
    # Helper function to build a project & then move its dist
    # artifacts into a subfolder under build_root.
    # ------------------------------------------------------
    def build_and_collect_artifacts(project_path: str, build_root: str):
        # Run 'poetry build' inside project_path
        run_command(["poetry", "build"], cwd=project_path)

        # dist/ subfolder inside that project
        dist_folder = os.path.join(project_path, "dist")
        if not os.path.isdir(dist_folder):
            return  # Might be an error or something, but let's just skip.

        # For neatness, name the subfolder after the project itself:
        project_name = os.path.basename(os.path.normpath(project_path))
        # e.g. build_root/mypackage
        target_folder = os.path.join(build_root, project_name)
        os.makedirs(target_folder, exist_ok=True)

        # Move any wheels/sdists to the target_folder
        for artifact in glob.glob(os.path.join(dist_folder, "*")):
            artifact_name = os.path.basename(artifact)
            dest_file = os.path.join(target_folder, artifact_name)
            print(f"Moving {artifact} -> {dest_file}")
            shutil.move(artifact, dest_file)

        # Optionally remove the now-empty dist folder:
        shutil.rmtree(dist_folder, ignore_errors=True)

    # ------------------------------------------------------
    # Decide how to create or re-use build_root
    # ------------------------------------------------------
    created_temp_dir = False
    if build_dir:
        build_root = os.path.abspath(build_dir)
        os.makedirs(build_root, exist_ok=True)
        print(f"Using user-specified build directory: {build_root}")
    else:
        build_root = tempfile.mkdtemp(prefix="mono_build_")
        created_temp_dir = True
        print(f"Using temporary directory for clones: {build_root}")

    # ------------------------------------------------------
    # 1) Build Local Path Dependencies
    # ------------------------------------------------------
    path_deps = extract_path_dependencies(aggregator_path)
    if path_deps:
        print(f"\nFound {len(path_deps)} local path dependencies. Building each ...")
        for path_rel in path_deps:
            sub_path = os.path.join(aggregator_dir, path_rel)
            pyproj_file = os.path.join(sub_path, "pyproject.toml")

            if not os.path.isdir(sub_path):
                print(f"Skipping '{path_rel}': directory not found at {sub_path}.")
                continue
            if not os.path.isfile(pyproj_file):
                print(f"Skipping '{sub_path}': no pyproject.toml found.")
                continue

            print(f"Building local path dependency in {sub_path} ...")
            try:
                build_and_collect_artifacts(sub_path, build_root)
            except Exception as e:
                print(f"Failed to build local path dependency {sub_path}: {e}")
    else:
        print("No local path dependencies found to build.")

    # ------------------------------------------------------
    # 2) Build Git Dependencies
    # ------------------------------------------------------
    git_deps = extract_git_dependencies(aggregator_path)
    if not git_deps:
        print("No Git dependencies found. Done.")
        # Possibly remove the build_root if it's temp and cleanup is True
        if created_temp_dir and cleanup:
            shutil.rmtree(build_root, ignore_errors=True)
            print(f"Cleaned up temporary directory: {build_root}")
        else:
            if not created_temp_dir:
                print(f"User-specified build directory left intact: {build_root}")
            else:
                print(f"Temporary build directory not removed: {build_root} (cleanup=False).")
        return

    print(f"\nFound {len(git_deps)} Git dependencies to build.")
    grouped_deps: Dict[Tuple[str, str], List[Tuple[str, str]]] = {}

    for dep_name, details in git_deps.items():
        git_url = details.get("git")
        branch = details.get("branch", "main")
        subdir = details.get("subdirectory", ".")
        key = (git_url, branch)

        if key not in grouped_deps:
            grouped_deps[key] = []
        grouped_deps[key].append((dep_name, subdir))

    # Clone & build each grouping:
    try:
        for (git_url, branch), sub_pkgs in grouped_deps.items():
            print(f"\n---\nCloning {git_url} (branch: {branch}) to build {len(sub_pkgs)} sub-packages.")

            clone_dir_name = branch.replace("/", "-") + "_clone"
            clone_dir = os.path.join(build_root, clone_dir_name)

            clone_cmd = [
                "git", "clone",
                "--depth", "1",
                "--branch", branch,
                git_url,
                clone_dir
            ]

            try:
                run_command(clone_cmd)
            except Exception as e:
                print(f"Failed to clone {git_url}: {e}")
                continue

            # Build each subdirectory in the newly cloned repo
            for (dep_name, subdir) in sub_pkgs:
                sub_path = os.path.join(clone_dir, subdir)
                pyproj_file = os.path.join(sub_path, "pyproject.toml")

                if not os.path.isdir(sub_path):
                    print(f"[{dep_name}] Subdirectory '{subdir}' not found.")
                    continue
                if not os.path.isfile(pyproj_file):
                    print(f"[{dep_name}] No pyproject.toml in {sub_path}. Skipping.")
                    continue

                print(f"[{dep_name}] Building package in {sub_path}")
                try:
                    build_and_collect_artifacts(sub_path, build_root)
                except Exception as e:
                    print(f"[{dep_name}] Failed to build: {e}")

        print("\nAll local path & Git sub-packages processed.")
    finally:
        if created_temp_dir and cleanup:
            shutil.rmtree(build_root, ignore_errors=True)
            print(f"Cleaned up temporary directory: {build_root}")
        else:
            if not created_temp_dir:
                print(f"User-specified build directory left intact: {build_root}")
            else:
                print(f"Temporary build directory not removed: {build_root} (cleanup=False).")


def poetry_publish(
    file: str = None,
    directory: str = None,
    recursive: bool = False,
    username: str = None,
    password: str = None
):
    """
    Builds and publishes packages to PyPI (or another configured repository).

    You can specify one of:
      - file=<path to pyproject.toml>
      - directory=<path to directory containing pyproject.toml>
      - directory=<path> + recursive=True to walk subdirectories

    :param file: Explicit path to a pyproject.toml file.
    :param directory: Path to a directory that has (or contains) a pyproject.toml.
    :param recursive: If True, search subdirectories for pyproject.toml files.
    :param username: PyPI username (optional). If not provided, rely on Poetry config or interactive prompt.
    :param password: PyPI password (optional).
    """

    username = username or os.environ.get("POETRY_PYPI_USERNAME")
    password = password or os.environ.get("POETRY_PYPI_PASSWORD")

    try:
        pyproject_files = find_pyproject_files(file, directory, recursive)
    except Exception as e:
        print(f"Error discovering pyproject files: {e}", file=sys.stderr)
        sys.exit(1)

    for pyproj in pyproject_files:
        project_dir = os.path.dirname(pyproj)
        print(f"Publishing package from {project_dir} ...")

        # 1. Build the package
        try:
            run_command(["poetry", "build"], cwd=project_dir)
        except Exception as e:
            print(f"Failed to build package in {project_dir}: {e}", file=sys.stderr)
            sys.exit(1)

        # 2. Publish the package
        publish_cmd = ["poetry", "publish"]
        # If you want verbose output, you can add ["-vv"] or similar
        if username and password:
            publish_cmd.extend(["--username", username, "--password", password])
        # If you have a custom repository name, you can also add --repository <repo>

        try:
            run_command(publish_cmd, cwd=project_dir)
        except Exception as e:
            print(f"Failed to publish package in {project_dir}: {e}", file=sys.stderr)
            sys.exit(1)


#!/usr/bin/env python3
"""
test_ops.py

Implements new test modes for a monorepo:
  - run_monorepo_tests: one unified pytest run
  - run_subpackage_tests: run pytest in each subpackage
  - (single-package tests are already handled by poetry_ops.run_pytests)
"""

import os
import sys
import glob
import subprocess
from typing import List

from .poetry_ops import run_pytests   # Your existing function
from .pyproject_ops import find_pyproject_files, extract_path_dependencies
# or from .pyproject_ops import ... if you prefer

def run_monorepo_tests(root_dir: str, num_workers: int = 1):
    """
    Runs a single pytest invocation at the top level.

    This typically discovers tests in all subdirectories automatically,
    assuming standard pytest discovery rules.
    """
    print(f"[monorepo] Running one pytest run from {root_dir} ...")
    run_pytests(test_directory=root_dir, num_workers=num_workers)

def run_subpackage_tests(root_dir: str, num_workers: int = 1):
    """
    Discovers subpackages and runs a separate pytest invocation in each.
    If there's an aggregator pyproject.toml with path dependencies, we can use that
    for subpackage discovery. Otherwise, do something like find all 'pyproject.toml' 
    that have 'package-mode=true', etc.
    """
    print(f"[subpackages] Searching for subpackages in {root_dir} ...")

    # 1) Possibly look for an aggregator in the specified directory
    #    If none found, fallback to a recursive search for "pyproject.toml"
    #    that have 'package-mode' or contain tests, etc.
    try:
        aggregator_files = find_pyproject_files(
            file=None, directory=root_dir, recursive=False
        )
        aggregator_pyproject = aggregator_files[0] if aggregator_files else None
    except Exception:
        aggregator_pyproject = None

    subpackage_dirs: List[str] = []

    if aggregator_pyproject:
        print(f"Aggregator file found: {aggregator_pyproject}")
        # Extract local path deps from aggregator
        path_deps = extract_path_dependencies(aggregator_pyproject)
        aggregator_dir = os.path.dirname(aggregator_pyproject)

        for p in path_deps:
            full_subpkg_path = os.path.join(aggregator_dir, p)
            pyproj = os.path.join(full_subpkg_path, "pyproject.toml")
            if os.path.isdir(full_subpkg_path) and os.path.isfile(pyproj):
                subpackage_dirs.append(full_subpkg_path)
    else:
        # If no aggregator, do a fallback: search all pyproject.toml in subdirs
        # that are not the root, and treat them as subpackages
        # (e.g. skip aggregator or non-package-mode).
        all_pyprojects = find_pyproject_files(
            file=None, directory=root_dir, recursive=True
        )
        root_abs = os.path.abspath(root_dir)

        for pyproj in all_pyprojects:
            if os.path.dirname(pyproj) == root_abs:
                # This is the same dir as root_dir, skip aggregator if any
                continue
            subpackage_dirs.append(os.path.dirname(pyproj))

    if not subpackage_dirs:
        print("No subpackages found. Exiting.")
        return

    print(f"\nFound {len(subpackage_dirs)} subpackages. Testing each individually ...")

    failure_count = 0
    for sub_dir in subpackage_dirs:
        print(f"\n=== Testing subpackage: {sub_dir} ===")
        try:
            run_pytests(test_directory=sub_dir, num_workers=num_workers)
        except SystemExit as e:
            # If your run_pytests() might do sys.exit(...) on errors,
            # handle that here:
            failure_count += 1
            print(f"Tests FAILED in {sub_dir}", file=sys.stderr)

    if failure_count > 0:
        print(f"\nFAIL: {failure_count} subpackages had test failures.")
        # You may prefer sys.exit(1) here:
        sys.exit(1)
    else:
        print("\nAll subpackages passed successfully!")
